/**
 * --------------------------------------------------------------------------
 * libretime-player (0.1.0): a player for a LibreTime webradio
 * Licensed under AGPL-3.0+
 * --------------------------------------------------------------------------
 */

import { Howl } from 'howler';
import request from 'superagent';

/**
 * ------------------------------------------------------------------------
 * Constants
 * ------------------------------------------------------------------------
 */

const Default = {
  /**
   * Ordered list of audio streams to play. The stream can be a string with
   * the URL if the format can be guessed from it. Otherwise, it must be an
   * object with 'src' and 'format' properties.
   * @option
   * @required
   * @type {Array|string}
   */
  streams: [],
  /**
   * URL to the LibreTime API to fetch live information, such as what is
   * currently playing and what is next. It is usually of the form of
   * `https://radio.example.org/api/live-info/`.
   * @option
   * @required
   * @type {string}
   */
  apiUrl: '',
  /**
   * Whether the player should start automatically at initialization.
   * @option
   * @type {boolean}
   */
  autoplay: false,
  /**
   * Timeout in milliseconds between metadata refresh from the API.
   * @option
   * @type {number}
   */
  refreshTimeout: 5000,
  /**
   * Status text displayed when the stream is loading.
   * @option
   * @type {string}
   */
  loadingText: 'Chargement&hellip;',
  /**
   * Status text displayed when the radio is on air.
   * @option
   * @type {string}
   */
  onAirText: 'À l\'antenne',
  /**
   * Status text displayed when the radio is offline.
   * @option
   * @type {string}
   */
  offlineText: 'Hors antenne',
  /**
   * Text displayed in the schedule list when nothing is planned next.
   * @option
   * @type {string}
   */
  emptyScheduleText: 'Rien de programmé',
  /**
   * Error message displayed when the stream could not be loaded.
   * @option
   * @type {string}
   */
  loadErrorMessage: 'Erreur : impossible de charger le flux.',
  /**
   * Error message displayed when the stream could not be played.
   * @option
   * @type {string}
   */
  playErrorMessage: 'Erreur : impossible de lire le flux.'
};

const ClassName = {
  PLAY: 'play',
  STOP: 'stop'
};

const Selector = {
  STATUS: '.now_playing',
  SCHEDULE_LIST: '.schedule_list',
  TOGGLER: '.airtime_button'
};

/**
 * ------------------------------------------------------------------------
 * Method Helpers
 * ------------------------------------------------------------------------
 */

const ISO_DATE = /^\s*\d{4}-\d\d-\d\d([T ]\d)?/;
const ISO_TZONE_RE = /(?:(Z)|([-+])(\d\d)(?::(\d\d))?)$/;

function isValidDate(date) {
  return !isNaN(date.valueOf());
}

function parseDate(str) {
  let timeZoneOffset = null;
  let isTimeUnspecified = false;
  const m = ISO_DATE.exec(str);

  if (m) {
    isTimeUnspecified = !m[1];

    if (isTimeUnspecified) {
      str += 'T00:00:00Z';
    } else {
      str = str.replace(ISO_TZONE_RE, (whole, z, sign, minutes, seconds) => {
        if (z) {
          timeZoneOffset = 0;
        } else {
          timeZoneOffset = (
            parseInt(minutes, 10) * 60 +
            parseInt(seconds || 0, 10)
          ) * (sign === '-' ? -1 : 1);
        }

        return '';
      }) + 'Z'; // otherwise will parse in local
    }
  }

  let date = new Date(str);

  if (!isValidDate(date)) {
    return null;
  }

  if (timeZoneOffset !== null) {
    date = new Date(date.valueOf() - timeZoneOffset * 60 * 1000);
  }

  return date;
}

function getRelativeDateString(date) {
  if (date instanceof Date && !isNaN(date)) {
    const today = new Date().toDateString();
    const string = date.toLocaleTimeString().replace(/:00$/, '');

    if (today !== date.toDateString()) {
      return `${date.toLocaleDateString()} ${string}`;
    }

    return string;
  }

  return '';
}

/**
 * ------------------------------------------------------------------------
 * Class Definition
 * ------------------------------------------------------------------------
 */

class RadioPlayer {
  constructor(element, options) {
    this.element = element;

    this._html = {};
    this._options = this._getOptions(options);

    this._audioUnlocked = true;
    this._hasErrorMessage = false;

    this._html.status = this.element.querySelector(Selector.STATUS);
    this._html.scheduleList = this.element.querySelector(Selector.SCHEDULE_LIST);
    this._html.toggler = this.element.querySelector(Selector.TOGGLER);

    if (this._html.toggler) {
      this._html.toggler.addEventListener('click', this._handleToggle.bind(this));
      this._html.toggler.classList.remove(ClassName.STOP);
      this._html.toggler.classList.add(ClassName.PLAY);
    }

    if (this._html.status || this._html.scheduleList) {
      if (this._html.status) {
        this._html.status.innerHTML = `<small>${this._options.loadingText}</small>`;
      }
      if (this._html.scheduleList) {
        this._html.scheduleList.innerHTML = `<li>${this._options.loadingText}</li>`;
      }

      this.refresh();
    }

    if (this._options.autoplay) {
      this.play();
    }
  }

  // Getters

  static get Default() {
    return Default;
  }

  // Public

  play() {
    if (!this._player) {
      this._load();
    }

    if (!this._player.playing()) {
      this._player.play();
    }
  }

  stop() {
    if (this._player) {
      this._player.unload();
    }
  }

  toggle() {
    if (this._player && this._player.playing()) {
      this.stop();
    } else {
      this.play();
    }
  }

  refresh() {
    request
      .get(this._options.apiUrl)
      .send({ type: 'interval', limit: '5' })
      .then((res) => {
        this._updateStatus(res.body);

        setTimeout(this.refresh.bind(this), this._options.refreshTimeout);
      });
  }

  // Private

  _getOptions(options) {
    options = {
      ...this.constructor.Default,
      ...typeof options === 'object' && options ? options : {}
    };

    if (!options.streams || options.streams.length === 0) {
      throw new Error('At least one stream must be defined');
    } else if (typeof options.streams === 'string') {
      options.streams = [options.streams];
    }

    if (!options.apiUrl) {
      throw new Error('The API URL to LibreTime must be defined');
    }

    return options;
  }

  _load() {
    if (this._player) {
      return;
    }

    const options = {
      src: [],
      formats: [],
      html5: true, // A live stream can only be played through HTML5 Audio.
      onplay: this._handlePlay.bind(this),
      onloaderror: this._handleLoadError.bind(this),
      onplayerror: this._handlePlayError.bind(this)
    };

    this._options.streams.forEach((stream) => {
      if (typeof stream === 'string') {
        options.src.push(stream);
        options.formats.push(null);
      } else {
        options.src.push(stream.src);
        options.formats.push(stream.format || null);
      }
    });

    this._player = new Howl(options);

    if (this._html.toggler) {
      this._html.toggler.setAttribute('disabled', 'true');

      this._player.on('play', () => {
        this._html.toggler.classList.remove(ClassName.PLAY);
        this._html.toggler.classList.add(ClassName.STOP);
        this._html.toggler.removeAttribute('disabled');
      }).on('stop', () => {
        this._html.toggler.classList.remove(ClassName.STOP);
        this._html.toggler.classList.add(ClassName.PLAY);
        this._html.toggler.removeAttribute('disabled');
      });
    }
  }

  _handlePlay() {
    this._audioUnlocked = true;
    this._hasErrorMessage = false;
  }

  _handleToggle() {
    // Audio will be soon unlocked or the stream is being loaded.
    if (this._audioUnlocked !== true) {
      return;
    }

    this.toggle();
  }

  _handleLoadError(id, err) {
    console.warn('loaderror:', err);

    this._hasErrorMessage = true;

    if (this._html.status) {
      this._html.status.innerHTML = this._options.loadErrorMessage;
    }

    if (this._html.toggler) {
      this._html.toggler.removeAttribute('disabled');
    }
  }

  _handlePlayError(id, err) {
    console.warn('playerror:', err);

    if (this._options.autoplay) {
      this._audioUnlocked = false;

      // This error can occur when audio is locked due to the lack of user
      // interaction. We try to play the stream as soon as it is unlocked.
      this._player.once('unlock', () => {
        if (this._audioUnlocked !== false) {
          return;
        }

        this._audioUnlocked = this._player.play();
      });
    } else {
      this._hasErrorMessage = true;

      if (this._html.status) {
        this._html.status.innerHTML = this._options.playErrorMessage;
      }
    }

    if (this._html.toggler) {
      this._html.toggler.removeAttribute('disabled');
    }
  }

  _updateStatus(data) {
    if (this._html.status && !this._hasErrorMessage) {
      let content;

      if (!data.current || !data.current.name) {
        if (data.currentShow && data.currentShow.length > 0) {
          content = `${this._options.onAirText}
            <span>${data.currentShow[0].name}</span>`;
        } else {
          content = this._options.offlineText;
        }
      } else {
        content = data.current.name;
      }

      this._html.status.innerHTML = content;
    }

    if (this._html.scheduleList) {
      let content;

      if (!data.next || !data.next.name) {
        content = this._options.emptyScheduleText;
      } else {
        content = data.next.name;

        const starts = parseDate(data.next.starts);
        const ends = parseDate(data.next.ends);

        if (starts) {
          content += `<span>${getRelativeDateString(starts)}`;
          if (ends) {
            content += ` - ${getRelativeDateString(ends)}`;
          }
          content += '</span>';
        }
      }

      this._html.scheduleList.innerHTML = `<li>${content}</li>`;
    }
  }
}

export default RadioPlayer;
